/* 
 * This file is part of the dcload Dreamcast serial loader
 *
 * Copyright (C) 2001 Andrew Kieschnick <andrewk@napalm-x.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/********************************************************
 * Win32 port and transmission fixes by NickK 8/22/2002 *
 ********************************************************/

#include <sys/utime.h>
#include "dc-tool.h"

/* The directory entry structure that KOS uses */

typedef struct
{
    long           nodeNumber;   /* INode Number */
    off_t          offset;       /* Offset to the next directory entry */
    unsigned short recordLength; /* The length of this record */
    unsigned char  type;         /* The type of file */
    char           name[256];    /* The name of the file */

    /* This is for the program's use only */
    HANDLE          fileList;
    WIN32_FIND_DATA fileData;
    BIDI_LIST_DATA  *listData; /* The position of this structure in the
                                  directory entry list */
} DIR_ENTRY;

/* Make sure the KOS will be getting correct data from the file
   stat structure */

#if S_IFDIR != 0040000
    #error S_IFDIR does not equal 0040000! Please correct the code to make the correct conversions of the file stat flags!
#endif

/* Use this to keep track of allocated and opened directory
   structures. Note: The file I/O doesn't need to be watched as
   closely since there's a closeall function :) */

BIDI_LIST *dirEntryList = NULL;

/********************************
 * Directory listing management *
 ********************************/

void dirEntryDestroyCallback(DIR_ENTRY *data)
{
     if (data != NULL)
     {
        if (data->fileList != NULL)
           FindClose(data->fileList);

        free(data);
     }
}

void dcLoadSysCallsInit(void)
{
     dirEntryList = BiDiListCreate();
}

void dcLoadSysCallsShutdown(void)
{
     if (dirEntryList != NULL)
     {
        BiDiListDestroy(dirEntryList, dirEntryDestroyCallback);
        dirEntryList = NULL;
     }
}

/*    do
    {
        printf("%s (%s)\n", fileData.cFileName,
                            (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? "Directory" : "File");
    } while (FindNextFile(fileList, &fileData));

    FindClose(fileList);*/

DIR_ENTRY *startDirectoryList(char *path)
{
    DIR_ENTRY     *dirEntry = NULL;
    char          *pathName = NULL;
    unsigned long offset;

    if (strlen(path) + 5 >= MAX_PATH) /* This path name is too long */
       return NULL;

    offset = strlen(path);

    if (offset <= 0) /* It could happen.. */
       return NULL;

    dirEntry = (DIR_ENTRY *)malloc(sizeof(DIR_ENTRY));
    if (!dirEntry)
       return NULL;

    dirEntry->nodeNumber = -1;
    memset(dirEntry->name, 0, 256);
    dirEntry->fileList = NULL;
    dirEntry->listData = NULL;

    pathName = (char *)malloc(strlen(path) + 5);
    if (!pathName)
    {
       free(dirEntry);
       return NULL;
    }

    memcpy(pathName, path, strlen(path));

    if (pathName[offset - 1] != '/' && pathName[offset - 1] != '\\')
    {
       /* Need to insert a '/' character at the end of the path
          name */
       pathName[offset] = '/';
       offset++;
    }

    pathName[offset]     = '*';
    pathName[offset + 1] = '.';
    pathName[offset + 2] = '*';
    pathName[offset + 3] = 0;

    /* Open a file searching handle and get the info for the first
       entry in the directory */
    dirEntry->fileList = FindFirstFile(pathName, &dirEntry->fileData);
    if (!dirEntry->fileList)
    {
       free(pathName);
       free(dirEntry);
       return NULL;
    }

    free(pathName);

    /* Add this new structure to the directory entry list to track
       it */
    BiDiListAdd(dirEntryList, dirEntry);

    /* Get the list data for this entry */
    dirEntry->listData = dirEntryList->listEnd;

    return dirEntry;
}

/*************************
 * System call functions *
 *************************/

void dcFStat(HANDLE hCommPort)
{
    int         fileHandle;
    struct stat filestat;
    unsigned int retval;

    recvUInt(hCommPort, (unsigned int *)&fileHandle);
    retval = fstat(fileHandle, &filestat);

    sendUInt(hCommPort, filestat.st_dev);
    sendUInt(hCommPort, filestat.st_ino);
    sendUInt(hCommPort, filestat.st_mode);
    sendUInt(hCommPort, filestat.st_nlink);
    sendUInt(hCommPort, filestat.st_uid);
    sendUInt(hCommPort, filestat.st_gid);
    sendUInt(hCommPort, filestat.st_rdev);
    sendUInt(hCommPort, filestat.st_size);
    sendUInt(hCommPort, 0); //filestat.st_blksize);
    sendUInt(hCommPort, 0); //filestat.st_blocks);
    sendUInt(hCommPort, filestat.st_atime);
    sendUInt(hCommPort, filestat.st_mtime);
    sendUInt(hCommPort, filestat.st_ctime);
    sendUInt(hCommPort, retval);
}

void dcWrite(HANDLE hCommPort)
{
    int           fileHandle;
    unsigned char *data;
    int           retval;
    int           count;

    recvUInt(hCommPort, (unsigned int *)&fileHandle);
    recvUInt(hCommPort, (unsigned int *)&count);

    data = malloc(count);
    recvData(hCommPort, data, count, 0);

    retval = _write(fileHandle, data, count);

    if (retval < count)
       printf("Write error: %d (count = %d) ", retval, count);

    sendUInt(hCommPort, retval);

    free(data);
}

void dcRead(HANDLE hCommPort)
{
    int           fileHandle;
    unsigned char *data;
    int           retval;
    int           count;

    recvUInt(hCommPort, (unsigned int *)&fileHandle);
    recvUInt(hCommPort, (unsigned int *)&count);

    data = malloc(count);

    retval = _read(fileHandle, data, count);

    if (retval < count)
       printf("Read error: %d (count = %d) ", retval, count);

    sendData(hCommPort, data, count, 0);
    sendUInt(hCommPort, retval);

    free(data);
}

void dcOpen(HANDLE hCommPort)
{
    unsigned int namelen;
    unsigned char *pathname;
    unsigned int flags;
    unsigned int fileFlags = 0;
    unsigned int mode;
    int          retval;

    recvUInt(hCommPort, &namelen);

    pathname = malloc(namelen);

    recvData(hCommPort, pathname, namelen, 0);

    recvUInt(hCommPort, &flags);
    recvUInt(hCommPort, &mode);

/*
    Byte 1 -> File access mode
        0  - Read only
        1  - Write only
        2  - Read and write
        10 - Append

    Byte 2 -> Modifiers (use the following values as masks)
        0x200 - Create a new file if it doesn't exist but do not
                truncate unless 0x400 is also present
        0x400 - Truncate or create a new file
*/

    switch (flags & 0xFF)
    {
           case 0: /* Read only */
                fileFlags = _O_RDONLY;
           break;

           case 1: /* Write only */
                fileFlags = _O_WRONLY;
           break;

           case 2: /* Read and write */
                fileFlags = _O_RDWR;
           break;

           case 10: /* Append */
                fileFlags = _O_APPEND;
           break;
    }

    if (flags & 0x200) /* Create new */
       fileFlags |= _O_CREAT;

    if (flags & 0x400) /* Truncate */
       fileFlags |= _O_TRUNC;

    fileFlags |= _O_BINARY;

    if (flags & 0x200)
       retval = _open(pathname, fileFlags, _S_IREAD | _S_IWRITE);
       else
       retval = _open(pathname, fileFlags);

    if (retval==-1)
       printf("Could not open the file '%s'!\n", pathname);

    sendUInt(hCommPort, retval);

    free(pathname);
}

void dcClose(HANDLE hCommPort)
{
     int fileHandle;
     int retVal;

     recvUInt(hCommPort, (unsigned int *)&fileHandle);

     retVal = _close(fileHandle);

     sendUInt(hCommPort, retVal);
}

void dcCreate(HANDLE hCommPort)
{
     unsigned int  namelen;
     unsigned char *pathname;
     int           retval;
     unsigned int  mode;

     recvUInt(hCommPort, &namelen);

     pathname = malloc(namelen);

     recvData(hCommPort, pathname, namelen, 0);
     recvUInt(hCommPort, &mode);

     if (CreateDirectory(pathname, NULL))
        retval = 0;
        else
        retval = -1;

     sendUInt(hCommPort, retval);

     free(pathname);
}

void dcLink(HANDLE hCommPort)
{
    unsigned int  namelen1, namelen2;
    unsigned char *pathname1, *pathname2;
    int           retval = 0;

    recvUInt(hCommPort, &namelen1);
    pathname1 = malloc(namelen1);

    recvData(hCommPort, pathname1, namelen1, 0);

    recvUInt(hCommPort, &namelen2);
    pathname2 = malloc(namelen2);

    recvData(hCommPort, pathname2, namelen2, 0);

    if (!CopyFile(pathname1, pathname2, TRUE))
       retval = -1;

    sendUInt(hCommPort, retval);

    free(pathname1);
    free(pathname2);
}

void dcUnlink(HANDLE hCommPort)
{
    unsigned int  namelen;
    unsigned char *pathname;
    int           retval = 0;

    recvUInt(hCommPort, &namelen);

    pathname = malloc(namelen);

    recvData(hCommPort, pathname, namelen, 0);

    if (!DeleteFile(pathname))
       retval = -1;

    sendUInt(hCommPort, retval);

    free(pathname);
}

void dcChdir(HANDLE hCommPort)
{
    unsigned int  namelen;
    unsigned char *pathname;
    int           retval;

    recvUInt(hCommPort, &namelen);

    pathname = malloc(namelen);

    recvData(hCommPort, pathname, namelen, 0);

    if (SetCurrentDirectory(pathname))
       retval = 0;
       else
       retval = -1;

    sendUInt(hCommPort, retval);

    free(pathname);
}

void dcChmod(HANDLE hCommPort)
{
    unsigned int  namelen;
    unsigned char *pathname;
    unsigned int  retval;
    unsigned int  mode;

    recvUInt(hCommPort, &namelen);

    pathname = malloc(namelen);

    recvData(hCommPort, pathname, namelen, 0);
    recvUInt(hCommPort, &mode);

    retval = 0; //chmod(pathname, mode);

    sendUInt(hCommPort, retval);

    free(pathname);
}

void dcLSeek(HANDLE hCommPort)
{
    int fileHandle;
    int offset;
    int mode;
    int retval;
    int seekMode = SEEK_CUR;

    recvUInt(hCommPort, (unsigned int *)&fileHandle);
    recvUInt(hCommPort, (unsigned int *)&offset);
    recvUInt(hCommPort, (unsigned int *)&mode);

    switch (mode)
    {
           case 0:
                seekMode = SEEK_SET;
           break;

           case 1:
                seekMode = SEEK_CUR;
           break;

           case 2:
                seekMode = SEEK_END;
           break;
    }

    retval = lseek(fileHandle, offset, seekMode);

    sendUInt(hCommPort, retval);
}

void dcTime(HANDLE hCommPort)
{
    time_t t;

    time(&t);

    sendUInt(hCommPort, t);
}

void dcStat(HANDLE hCommPort)
{
    unsigned int  namelen;
    unsigned char *filename;
    struct stat filestat;
    int retval;

    recvUInt(hCommPort, &namelen);

    filename = malloc(namelen);

    recvData(hCommPort, filename, namelen, 0);

    retval = stat(filename, &filestat);

    sendUInt(hCommPort, filestat.st_dev);
    sendUInt(hCommPort, filestat.st_ino);
    sendUInt(hCommPort, filestat.st_mode);
    sendUInt(hCommPort, filestat.st_nlink);
    sendUInt(hCommPort, filestat.st_uid);
    sendUInt(hCommPort, filestat.st_gid);
    sendUInt(hCommPort, filestat.st_rdev);
    sendUInt(hCommPort, filestat.st_size);
    sendUInt(hCommPort, 0);//filestat.st_blksize);
    sendUInt(hCommPort, 0);//filestat.st_blocks);
    sendUInt(hCommPort, filestat.st_atime);
    sendUInt(hCommPort, filestat.st_mtime);
    sendUInt(hCommPort, filestat.st_ctime);

    sendUInt(hCommPort, retval);

    free(filename);
}

void dcUTime(HANDLE hCommPort)
{
    unsigned char *pathname;
    unsigned int  namelen;
    struct utimbuf tbuf;
    unsigned int   temp;
    int            retval;

    recvUInt(hCommPort, &namelen);

    pathname = malloc(namelen);

    recvData(hCommPort, pathname, namelen, 0);
    recvUInt(hCommPort, &temp);

    if (temp)
    {
       recvUInt(hCommPort, (unsigned int *)&tbuf.actime);
       recvUInt(hCommPort, (unsigned int *)&tbuf.modtime);

       retval = utime(pathname, &tbuf);
    } else
       retval = utime(pathname, NULL);

    sendUInt(hCommPort, retval);

    free(pathname);
}

void dcOpenDir(HANDLE hCommPort)
{
    DIR_ENTRY     *dirEntry = NULL;
    unsigned char *dirname;
    int           namelen;

    recvUInt(hCommPort, (unsigned int *)&namelen);

    dirname = malloc(namelen);

    recvData(hCommPort, dirname, namelen, 0);

    dirEntry = startDirectoryList((char *)dirname);

    sendUInt(hCommPort, (unsigned int)dirEntry);

    free(dirname);
}

void dcCloseDir(HANDLE hCommPort)
{
    DIR_ENTRY *dirEntry = NULL;
    int       retVal    = 0;

    recvUInt(hCommPort, (unsigned int *)&dirEntry);

    if (dirEntry != NULL)
       BiDiListRemove(dirEntryList, dirEntry->listData,
                      dirEntryDestroyCallback);
       else
        retVal = -1;

    sendUInt(hCommPort, retVal);
}

void dcReadDir(HANDLE hCommPort)
{
     DIR_ENTRY *dirEntry = NULL;

     recvUInt(hCommPort, (unsigned int *)&dirEntry);

     if (dirEntry != NULL)
     {
        if ((dirEntry->nodeNumber >= 0 &&
            FindNextFile(dirEntry->fileList, &dirEntry->fileData)) ||
            dirEntry->nodeNumber < 0)
        {
           sendUInt(hCommPort, 1);
           sendUInt(hCommPort, dirEntry->nodeNumber + 1);
           sendUInt(hCommPort, 0); /* offset */
           sendUInt(hCommPort, 0); /* record length */
           sendUInt(hCommPort, 0); /* type */

           dirEntry->nodeNumber++;

           /* Copy the name that's contained in fileData into the name
              string in the DIR_ENTRY structure. If the length of the
              file name is greater than or equal to 255 then the name
              is set to all zeroes except for the first character
              which is set to '.' which identifies the current local
              directory */
           if (strlen(dirEntry->fileData.cFileName) >= 255)
           {
              memset(dirEntry->name, 0, 256);
              dirEntry->name[0] = '.';
           } else
              memcpy(dirEntry->name, dirEntry->fileData.cFileName,
                     strlen(dirEntry->fileData.cFileName) + 1);

           sendUInt(hCommPort, strlen(dirEntry->name) + 1);
           sendData(hCommPort, dirEntry->name,
                    strlen(dirEntry->name) + 1, 0);
        } else /* No more directory entries left */
           sendUInt(hCommPort, 0);
     } else
        sendUInt(hCommPort, 0);
}

void dcCDFSRedirReadSectors(HANDLE hCommPort, FILE *f)
{
    int           start;
    int           num;
    unsigned char *buf;

    recvUInt(hCommPort, (unsigned int *)&start);
    recvUInt(hCommPort, (unsigned int *)&num);

    start -= 150; 

    fseek(f, start << 11, SEEK_SET);

    buf = malloc(num << 11);

    fread(buf, 1, num << 11, f);

    sendData(hCommPort, buf, num << 11, 0);
}